from rest_framework import serializers
from pandas import DataFrame
from django.core.exceptions import ImproperlyConfigured
import datetime


class PandasSerializer(serializers.ListSerializer):
    """
    Transforms dataset into a dataframe and applies an index
    """
    read_only = True
    index_none_value = None
    wq_chart_type = None

    def get_index(self, dataframe):
        return self.get_index_fields()

    def get_dataframe(self, data):
        dataframe = DataFrame(data)
        index = self.get_index(dataframe)
        if index:
            if self.index_none_value is not None:
                for key in index:
                    try:
                        dataframe[key].fillna(
                            self.index_none_value, inplace=True
                        )
                    except ValueError:
                        pass
            dataframe.set_index(index, inplace=True)
        else:
            # Name auto-index column to ensure valid CSV output
            dataframe.index.name = 'row'
        return dataframe

    def transform_dataframe(self, dataframe):
        view = self.context.get('view', None)
        if view and hasattr(view, 'transform_dataframe'):
            return self.context['view'].transform_dataframe(dataframe)
        return dataframe

    @property
    def data(self):
        data = super(serializers.ListSerializer, self).data
        if isinstance(data, DataFrame) or data:
            dataframe = self.get_dataframe(data)
            return self.transform_dataframe(dataframe)
        else:
            return DataFrame([])

    def to_representation(self, data):
        if isinstance(data, DataFrame):
            return data
        return super(PandasSerializer, self).to_representation(data)

    @property
    def model_serializer(self):
        serializer = type(self.child)
        if serializer.__name__ == 'SerializerWithListSerializer':
            return serializer.__bases__[0]
        return serializer

    @property
    def model_serializer_meta(self):
        return getattr(self.model_serializer, 'Meta', object())

    def get_index_fields(self):
        """
        List of fields to use for index
        """
        index_fields = self.get_meta_option('index', [])
        if index_fields:
            return index_fields

        model = getattr(self.model_serializer_meta, 'model', None)
        if model:
            pk_name = model._meta.pk.name
            if pk_name in self.child.get_fields():
                return [pk_name]

        return []

    def get_meta_option(self, name, default=None):
        meta_name = 'pandas_' + name
        value = getattr(self.model_serializer_meta, meta_name, None)

        if value is None:
            if default is not None:
                return default
            else:
                raise ImproperlyConfigured(
                    "%s should be specified on %s.Meta" %
                    (meta_name, self.model_serializer.__name__)
                )
        return value


class PandasUnstackedSerializer(PandasSerializer):
    """
    Pivots dataframe so commonly-repeating values are across the top in a
    multi-row header.  Intended for use with e.g. time series data, where the
    header includes metadata applicable to each time series.
    (Use with wq/chart.js' timeSeries() function)
    """
    index_none_value = '-'
    wq_chart_type = 'timeSeries'

    def get_index(self, dataframe):
        """
        Include header fields in initial index for later unstacking
        """
        return self.get_index_fields() + self.get_header_fields()

    def transform_dataframe(self, dataframe):
        """
        Unstack the dataframe so header fields are across the top.
        """
        dataframe.columns.name = ""

        for i in range(len(self.get_header_fields())):
            dataframe = dataframe.unstack()

        # Remove blank rows / columns
        dataframe = dataframe.dropna(
            axis=0, how='all'
        ).dropna(
            axis=1, how='all'
        )
        return dataframe

    def get_header_fields(self):
        """
        Series metadata fields for header (first few rows)
        """
        return self.get_meta_option('unstacked_header')


class PandasScatterSerializer(PandasSerializer):
    """
    Pivots dataframe into a format suitable for plotting two series
    against each other as x vs y on a scatter plot.
    (Use with wq/chart.js' scatter() function)
    """
    index_none_value = '-'
    wq_chart_type = 'scatter'

    def get_index(self, dataframe):
        """
        Include scatter & header fields in initial index for later unstacking
        """
        return (
            self.get_index_fields() +
            self.get_header_fields() +
            self.get_coord_fields()
        )

    def transform_dataframe(self, dataframe):
        """
        Unstack the dataframe so header consists of a composite 'value' header
        plus any other header fields.
        """
        coord_fields = self.get_coord_fields()
        header_fields = self.get_header_fields()

        # Remove any pairs that don't have data for both x & y
        for i in range(len(coord_fields)):
            dataframe = dataframe.unstack()
        dataframe = dataframe.dropna(axis=1, how='all')
        dataframe = dataframe.dropna(axis=0, how='any')

        # Unstack series header
        for i in range(len(header_fields)):
            dataframe = dataframe.unstack()

        # Compute new column headers
        columns = []
        for i in range(len(header_fields) + 1):
            columns.append([])

        for col in dataframe.columns:
            value_name = col[0]
            coord_names = list(col[1:len(coord_fields) + 1])
            header_names = list(col[len(coord_fields) + 1:])
            coord_name = ''
            for name in coord_names:
                if name != self.index_none_value:
                    coord_name += name + '-'
            coord_name += value_name
            columns[0].append(coord_name)
            for i, header_name in enumerate(header_names):
                columns[1 + i].append(header_name)

        dataframe.columns = columns
        dataframe.columns.names = [''] + header_fields

        return dataframe

    def get_coord_fields(self):
        """
        Fields that will be collapsed into a single header with the name of
        each coordinate.
        """
        return self.get_meta_option('scatter_coord')

    def get_header_fields(self):
        """
        Other header fields, if any
        """
        return self.get_meta_option('scatter_header', [])


class PandasBoxplotSerializer(PandasSerializer):
    """
    Compute boxplot statistics on dataframe columns, creating a new unstacked
    dataframe where each row describes a boxplot.
    (Use with wq/chart.js' boxplot() function)
    """
    index_none_value = '-'
    wq_chart_type = 'boxplot'

    def get_index(self, dataframe):
        group_field = self.get_group_field()
        date_field = self.get_date_field()
        header_fields = self.get_header_fields()

        if date_field:
            group_fields = [date_field, group_field]
        else:
            group_fields = [group_field]
        return group_fields + header_fields

    def transform_dataframe(self, dataframe):
        """
        Use matplotlib to compute boxplot statistics on e.g. timeseries data.
        """
        grouping = self.get_grouping(dataframe)
        group_field = self.get_group_field()
        header_fields = self.get_header_fields()

        if "series" in grouping:
            # Unstack so each series is a column
            for i in range(len(header_fields) + 1):
                dataframe = dataframe.unstack()

        groups = {
            col: dataframe[col]
            for col in dataframe.columns
        }

        if "year" in grouping:
            interval = "year"
        elif "month" in grouping:
            interval = "month"
        else:
            interval = None

        # Compute stats for each column, potentially grouped by year
        all_stats = []
        for header, series in groups.items():
            if interval:
                series_stats = self.boxplots_for_interval(series, interval)
            else:
                interval = None
                series_stats = [self.compute_boxplot(series)]

            series_infos = []
            for series_stat in series_stats:
                series_info = {}
                if isinstance(header, tuple):
                    value_name = header[0]
                    col_values = header[1:]
                else:
                    value_name = header
                    col_values = []
                col_names = zip(dataframe.columns.names[1:], col_values)
                for col_name, value in col_names:
                    series_info[col_name] = value
                for stat_name, val in series_stat.items():
                    if stat_name == interval:
                        series_info[stat_name] = val
                    else:
                        series_info[value_name + '-' + stat_name] = val
                series_infos.append(series_info)
            all_stats += series_infos

        dataframe = DataFrame(all_stats)
        if 'series' in grouping:
            index = header_fields + [group_field]
            unstack = len(header_fields)
            if interval:
                index = [interval] + index
                unstack += 1
        else:
            index = [interval]
            unstack = 0

        dataframe.set_index(index, inplace=True)
        dataframe.columns.name = ''
        for i in range(unstack):
            dataframe = dataframe.unstack()

        # Remove blank columns
        dataframe = dataframe.dropna(axis=1, how='all')
        return dataframe

    def get_grouping(self, dataframe):
        request = self.context.get('request', None)
        datasets = len(dataframe.columns)
        if request:
            group = request.GET.get('group', None)
            if group:
                return group
        return default_grouping(datasets, self.get_date_field())

    def boxplots_for_interval(self, series, interval):
        def get_interval_name(date):
            if isinstance(date, tuple):
                date = date[0]
            if hasattr(date, 'count') and date.count('-') == 2:
                date = datetime.datetime.strptime(date, "%Y-%m-%d")
            return getattr(date, interval)

        interval_stats = []
        groups = series.groupby(get_interval_name).groups
        for interval_name, group in groups.items():
            stats = self.compute_boxplot(series[group])
            stats[interval] = interval_name
            interval_stats.append(stats)
        return interval_stats

    def compute_boxplot(self, series):
        """
        Compute boxplot for given pandas Series.
        """
        from matplotlib.cbook import boxplot_stats
        series = series[series.notnull()]
        if len(series.values) == 0:
            return {}
        stats = boxplot_stats(list(series.values))[0]
        stats['count'] = len(series.values)
        stats['fliers'] = "|".join(map(str, stats['fliers']))
        return stats

    def get_group_field(self):
        """
        Categorical field to group datasets by.
        """
        return self.get_meta_option('boxplot_group')

    def get_date_field(self):
        """
        Date field to group datasets by year or month.
        """
        return self.get_meta_option('boxplot_date', False)

    def get_header_fields(self):
        """
        Additional series metadata for boxplot column headers
        """
        return self.get_meta_option('boxplot_header', [])


class SimpleSerializer(serializers.Serializer):
    """
    Simple serializer for non-model (simple) views
    """
    def to_representation(self, obj):
        return obj


def default_grouping(datasets, date_field=None):
    """
    Heuristic for default boxplot grouping
    """
    if datasets > 20 and date_field:
        # Group all data by year
        return "year"
    elif datasets > 10 or not date_field:
        # Compare series but don't break down by year
        return "series"
    else:
        # 10 or fewer datasets, break down by both series and year
        return "series-year"



from .views import (
    PandasMixin,
    PandasSimpleView,
    PandasView,
    PandasViewSet,
)
from .serializers import (
    PandasSerializer,
    PandasUnstackedSerializer,
    PandasScatterSerializer,
    PandasBoxplotSerializer,
    SimpleSerializer,
)
from .renderers import (
    PandasBaseRenderer,
    PandasFileRenderer,
    PandasCSVRenderer,
    PandasTextRenderer,
    PandasJSONRenderer,
    PandasExcelRenderer,
    PandasOldExcelRenderer,
    PandasImageRenderer,
    PandasPNGRenderer,
    PandasSVGRenderer,
)


__all__ = [
    'PandasMixin',
    'PandasSimpleView',
    'PandasView',
    'PandasViewSet',

    'PandasSerializer',
    'PandasUnstackedSerializer',
    'PandasScatterSerializer',
    'PandasBoxplotSerializer',
    'SimpleSerializer',

    'PandasBaseRenderer',
    'PandasFileRenderer',
    'PandasCSVRenderer',
    'PandasTextRenderer',
    'PandasJSONRenderer',
    'PandasExcelRenderer',
    'PandasOldExcelRenderer',
    'PandasImageRenderer',
    'PandasPNGRenderer',
    'PandasSVGRenderer',
]



from rest_framework.renderers import BaseRenderer, TemplateHTMLRenderer
from rest_framework import status
from tempfile import mkstemp
from pandas import DataFrame

try:
    # Python 2 (uses str)
    from StringIO import StringIO
except ImportError:
    # Python 3 (Python 2 equivalent uses unicode)
    from io import StringIO

import os
from io import BytesIO


RESPONSE_ERROR = (
    "Response data is a %s, not a DataFrame! "
    "Did you extend PandasMixin?"
)


class PandasBaseRenderer(BaseRenderer):
    """
    Renders DataFrames using their built in pandas implementation.
    Only works with serializers that return DataFrames as their data object.
    Uses a StringIO to capture the output of dataframe.to_[format]()
    """
    def render(self, data, accepted_media_type=None, renderer_context=None):
        if 'response' in renderer_context:
            status_code = renderer_context['response'].status_code
            if not status.is_success(status_code):
                return "Error: %s" % data.get('detail', status_code)

        if not isinstance(data, DataFrame):
            raise Exception(
                RESPONSE_ERROR % type(data).__name__
            )

        name = getattr(self, 'function', "to_%s" % self.format)
        if not hasattr(data, name):
            raise Exception("Data frame is missing %s property!" % name)

        self.init_output()
        args = self.get_pandas_args(data)
        kwargs = self.get_pandas_kwargs(data, renderer_context)
        self.render_dataframe(data, name, *args, **kwargs)
        return self.get_output()

    def render_dataframe(self, data, name, *args, **kwargs):
        function = getattr(data, name)
        function(*args, **kwargs)

    def init_output(self):
        self.output = StringIO()

    def get_output(self):
        return self.output.getvalue()

    def get_pandas_args(self, data):
        return [self.output]

    def get_pandas_kwargs(self, data, renderer_context):
        return {}


class PandasFileRenderer(PandasBaseRenderer):
    """
    Renderer for output formats that absolutely must use a file (i.e. Excel)
    """
    def init_output(self):
        file, filename = mkstemp(suffix='.' + self.format)
        self.filename = filename
        os.close(file)

    def get_pandas_args(self, data):
        return [self.filename]

    def get_output(self):
        file = open(self.filename, 'rb')
        result = file.read()
        file.close()
        os.unlink(self.filename)
        return result


class PandasHTMLRenderer(TemplateHTMLRenderer, PandasBaseRenderer):
    media_type = "text/html"
    format = "html"

    def render(self, data, accepted_media_type=None, renderer_context=None):
        table = PandasBaseRenderer.render(
            self, data, accepted_media_type, renderer_context,
        )

        return TemplateHTMLRenderer.render(
            self, {'table': table}, accepted_media_type, renderer_context,
        )

    def get_template_context(self, data, renderer_context):
        view = renderer_context['view']
        request = renderer_context['request']

        data['name'] = view.get_view_name()
        data['description'] = view.get_view_description(html=True)
        data['url'] = request.path.replace('.html', '')
        full_path = request.get_full_path()
        if '?' in full_path:
            data['url_params'] = full_path[full_path.index('?'):]
        data['available_formats'] = [
            cls.format for cls in view.renderer_classes
            if cls.format != 'html'
        ]

        chart_type = view.pandas_serializer_class.wq_chart_type
        if chart_type:
            data['wq_chart_type'] = chart_type

        if hasattr(view, 'get_template_context'):
            data.update(view.get_template_context(data))

        return data

    def get_pandas_kwargs(self, data, renderer_context):
        return {
            'classes': 'ui-table table-stripe',
            'na_rep': '',
        }


class PandasCSVRenderer(PandasBaseRenderer):
    """
    Renders data frame as CSV
    """
    media_type = "text/csv"
    format = "csv"

    def get_pandas_kwargs(self, data, renderer_context):
        return {'encoding': self.charset}


class PandasTextRenderer(PandasCSVRenderer):
    """
    Renders data frame as CSV, but uses text/plain as media type
    """
    media_type = "text/plain"
    format = "txt"
    function = "to_csv"


class PandasJSONRenderer(PandasBaseRenderer):
    """
    Renders data frame as JSON
    """
    media_type = "application/json"
    format = "json"

    orient_choices = {
        'records-index',  # Unique to DRP
        'split',
        'records',
        'index',
        'columns',
        'values',
        'table',
    }
    default_orient = 'records-index'

    date_format_choices = {'epoch', 'iso'}
    default_date_format = 'iso'

    def get_pandas_kwargs(self, data, renderer_context):
        request = renderer_context['request']

        orient = request.GET.get('orient', '')
        if orient not in self.orient_choices:
            orient = self.default_orient

        date_format = request.GET.get('date_format', '')
        if date_format not in self.date_format_choices:
            date_format = self.default_date_format

        return {
            'orient': orient,
            'date_format': date_format,
        }

    def render_dataframe(self, data, name, *args, **kwargs):
        if kwargs.get('orient') == 'records-index':
            kwargs['orient'] = 'records'
            data.reset_index(inplace=True)
        return super(PandasJSONRenderer, self).render_dataframe(
            data, name, *args, **kwargs
        )


class PandasExcelRenderer(PandasFileRenderer):
    """
    Renders data frame as Excel (.xlsx)
    """
    media_type = "application/vnd.openxmlformats-officedocument.spreadsheetml.sheet"  # noqa
    format = "xlsx"
    function = "to_excel"


class PandasOldExcelRenderer(PandasFileRenderer):
    """
    Renders data frame as Excel (.xls)
    """
    media_type = "application/vnd.ms-excel"
    format = "xls"
    function = "to_excel"


class PandasImageRenderer(PandasBaseRenderer):
    """
    Renders dataframe using built-in plot() function
    """
    function = "plot"
    matplotlib_backend = 'Agg'

    def init_output(self):
        import matplotlib
        matplotlib.use(self.matplotlib_backend)
        import matplotlib.pyplot as plt
        self.fig = plt.figure()
        self.ax = self.fig.add_subplot(111)

    def get_pandas_args(self, data):
        return []

    def get_pandas_kwargs(self, data, renderer_context):
        return {'ax': self.ax}

    def get_output(self):
        data = BytesIO()
        self.fig.savefig(data, format=self.format)
        return data.getvalue()


class PandasPNGRenderer(PandasImageRenderer):
    media_type = "image/png"
    format = "png"


class PandasSVGRenderer(PandasImageRenderer):
    media_type = "image/svg"
    format = "svg"



from .renderers import PandasBaseRenderer
try:
    from rest_framework.views import APIView
except ImportError as e:
    if 'APIView' in e.msg:
        raise ImportError(
            "Try importing rest_pandas before rest_framework.views"
        )
    else:
        raise
from rest_framework.generics import ListAPIView
from rest_framework.mixins import ListModelMixin
from rest_framework.viewsets import GenericViewSet
from rest_framework.response import Response

from django.conf import settings
from rest_framework.settings import perform_import

from .serializers import (
    SimpleSerializer, PandasSerializer
)

DEFAULT_TEMPLATE = False
PANDAS_RENDERERS = getattr(settings, "PANDAS_RENDERERS", None)
if PANDAS_RENDERERS is None:
    PANDAS_RENDERERS = (
        "rest_pandas.renderers.PandasCSVRenderer",
        "rest_pandas.renderers.PandasTextRenderer",
        "rest_pandas.renderers.PandasJSONRenderer",
        "rest_pandas.renderers.PandasExcelRenderer",
        "rest_pandas.renderers.PandasOldExcelRenderer",
        "rest_pandas.renderers.PandasPNGRenderer",
        "rest_pandas.renderers.PandasSVGRenderer",
    )
    if "rest_pandas" in settings.INSTALLED_APPS:
        DEFAULT_TEMPLATE = True
        PANDAS_RENDERERS = (
            "rest_pandas.renderers.PandasHTMLRenderer",
        ) + PANDAS_RENDERERS

PANDAS_RENDERERS = perform_import(PANDAS_RENDERERS, "PANDAS_RENDERERS")


class PandasMixin(object):
    pandas_serializer_class = PandasSerializer

    def with_list_serializer(self, cls):
        meta = getattr(cls, 'Meta', object)
        if getattr(meta, 'list_serializer_class', None):
            return cls

        class SerializerWithListSerializer(cls):
            class Meta(meta):
                list_serializer_class = self.pandas_serializer_class

        return SerializerWithListSerializer

    def get_serializer_class(self):

        # c.f rest_framework.generics.GenericAPIView
        # (not using super() since this is a mixin class)
        assert self.serializer_class is not None, (
            "'%s' should either include a `serializer_class` attribute, "
            "or override the `get_serializer_class()` method."
            % self.__class__.__name__
        )

        renderer = self.request.accepted_renderer
        if hasattr(renderer, 'get_default_renderer'):
            # BrowsableAPIRenderer
            renderer = renderer.get_default_renderer(self)

        if isinstance(renderer, PandasBaseRenderer):
            return self.with_list_serializer(self.serializer_class)
        else:
            return self.serializer_class


class PandasViewBase(PandasMixin):
    renderer_classes = PANDAS_RENDERERS
    pagination_class = None
    if DEFAULT_TEMPLATE:
        template_name = 'rest_pandas.html'


class PandasSimpleView(PandasViewBase, APIView):
    """
    Simple (non-model) Pandas API view; override get_data
    with a function that returns a list of dicts.
    """
    serializer_class = SimpleSerializer

    def get_data(self, request, *args, **kwargs):
        return []

    def get(self, request, *args, **kwargs):
        data = self.get_data(request, *args, **kwargs)
        serializer_class = self.get_serializer_class()
        serializer = serializer_class(data, many=True)
        return Response(serializer.data)


class PandasView(PandasViewBase, ListAPIView):
    """
    Pandas-capable model list view
    """
    pass


class PandasViewSet(PandasViewBase, ListModelMixin, GenericViewSet):
    """
    Pandas-capable model ViewSet (list only)
    """
    pass



import csv
from .renderers import StringIO


def parse_csv(string):
    """
    Rough port of wq/pandas.js to Python.  Useful for validating CSV output
    generated by Django REST Pandas.
    """
    if not string.startswith(','):
        data = []
        for row in csv.DictReader(StringIO(string)):
            for key, val in row.items():
                try:
                    row[key] = float(val)
                except ValueError:
                    pass
            data.append(row)
        return [{
            'data': data
        }]

    reader = csv.reader(StringIO(string))
    val_cols = None
    val_start = None
    id_cols = None
    for row in reader:
        if row[0] == '' and not val_cols:
            val_start = row.count('')
            val_cols = row[val_start:]
            col_meta = [{} for v in val_cols]
        elif row[-1] != '' and val_cols and not id_cols:
            key = row[0]
            for i, meta in enumerate(row[val_start:]):
                col_meta[i].update(**{key: meta})
        elif row[-1] == '' and not id_cols:
            id_cols = row[:row.index('')]
            meta_index = {}
            meta_i = 0
            datasets = []
            for i, ds1 in enumerate(col_meta):
                if i in meta_index:
                    continue
                meta_index[i] = meta_i
                meta_i += 1
                datasets.append(ds1)
                if i < len(col_meta):
                    for j, ds2 in enumerate(col_meta[i + 1:]):
                        if ds1 == ds2:
                            meta_index[i + j + 1] = i
            for d in datasets:
                d['data'] = []
        elif val_cols and id_cols:
            ids = {
                key: val
                for key, val in zip(id_cols, row[:len(id_cols)])
            }
            records = {}
            for i, val in enumerate(row[len(id_cols):]):
                mi = meta_index[i]
                if mi not in records:
                    data = ids.copy()
                else:
                    data = records[mi]
                try:
                    val = float(val)
                except ValueError:
                    pass
                if val != '':
                    data[val_cols[i]] = val
                    records[mi] = data
            for mi, data in records.items():
                datasets[mi]['data'].append(data)
    return datasets



import subprocess
from setuptools import setup, find_packages

LONG_DESCRIPTION = (
    "Serves up pandas dataframes via the Django REST Framework for client-side"
    "(i.e. d3.js) visualizations"
)


def parse_markdown_readme():
    """
    Convert README.md to RST via pandoc, and load into memory
    (fallback to LONG_DESCRIPTION on failure)
    """
    try:
        subprocess.call(
            ['pandoc', '-t', 'rst', '-o', 'README.rst', 'README.md']
        )
    except OSError:
        return LONG_DESCRIPTION

    # Attempt to load output
    try:
        readme = open('README.rst')
    except IOError:
        return LONG_DESCRIPTION
    else:
        return readme.read()

setup(
    name='rest-pandas',
    version='1.0.0',
    author='S. Andrew Sheppard',
    author_email='andrew@wq.io',
    url='https://github.com/wq/django-rest-pandas',
    license='MIT',
    packages=['rest_pandas'],
    package_data={
        'rest_pandas': [
            'mustache/*.*',
        ]
    },
    description=LONG_DESCRIPTION.strip(),
    long_description=parse_markdown_readme(),
    install_requires=[
        'djangorestframework>=3.3.1',
        'pandas>=0.19.0',
    ],
    classifiers=[
        'Framework :: Django',
        'Development Status :: 5 - Production/Stable',
        'Environment :: Web Environment',
        'License :: OSI Approved :: MIT License',
        'Natural Language :: English',
        'Programming Language :: Python :: 2',
        'Programming Language :: Python :: 2.7',
        'Programming Language :: Python :: 3',
        'Programming Language :: Python :: 3.4',
        'Programming Language :: Python :: 3.5',
        'Programming Language :: Python :: 3.6',
        'Framework :: Django',
        'Framework :: Django :: 1.8',
        'Framework :: Django :: 1.9',
        'Framework :: Django :: 1.10',
        'Framework :: Django :: 1.11',
        'Topic :: Scientific/Engineering :: Information Analysis',
        'Topic :: Scientific/Engineering :: Visualization',
    ],
    test_suite='tests',
    tests_require=[
        'wq.io', 'xlwt', 'openpyxl', 'matplotlib',
        'django', 'django-mustache'
    ],
)
